# This code has been developed under THALIS: MATENVMED project for Action 4.3 and co-financed by Greece and EU - NSRF 2007-2013.
# FPDE python module solves the flow potential problem for Hersonisos aquifer problem
from dolfin import *
import numpy as np
import scipy as sc
import matplotlib.pyplot as plt
from Wave import *

def FPDE(Q_k,KK,sh,flagplot,WX,WY,GW1X,GW1Y,GW2X,GW2Y,G1,G2,iflag):
    n=4
    phi = 128.125
    mesh = Mesh("H30.xml")
# Initialize mesh function for boundary domains
    boundaries = MeshFunction('size_t', mesh,"H30_facet_region.xml")
    subdomains = MeshFunction('size_t', mesh, "H30_physical_region.xml")
    parameters["reorder_dofs_serial"] = False

    coordinates = mesh.coordinates()
    sh=100.0
   # Delta function for each well
    class Delta(Expression):
        def __init__(self, eps):
            self.eps = eps
        def eval(self, values, x):
            eps = self.eps
            area=sh*sh
            if x[0]==WX[0] and x[1]==WY[0] :
                values[0] = eps-Q_k[0]/area
            elif x[0]==WX[1] and x[1]==WY[1] :
                values[0] = eps-Q_k[2]/area
            elif x[0]==WX[2] and x[1]==WY[2] :
                values[0] = eps-Q_k[2]/area
            elif x[0]==WX[3] and x[1]==WY[3] :
                values[0] = eps-Q_k[3]/area
            else:
                values[0] = eps



    V0 = FunctionSpace(mesh, 'DG', 0)
    kapa = Function(V0)
    kapa.vector()[:]=KK[:]

    V = FunctionSpace(mesh, "CG", 1)

# Define variational problem
    u = TrialFunction(V)
    v = TestFunction(V)

#  Neumann condition values
    g_n2=Constant("0.05")
    #g_n3=Constant("1.04")/1.65
    g_n3=Constant("0.4")
    g_n4=Constant("0.22")
    g_n6=Constant("0.355")
    g_n7=Constant("0.05")

    ds = Measure("ds")[boundaries]
    bcs = [DirichletBC(V, 0.0, boundaries, 10)]

    delta= Delta(0.0)
    a = inner(kapa*grad(u), grad(v))*dx
    L = inner(delta, v)*dx + g_n2 *v*ds(2)+g_n3 *v*ds(3)+g_n4 *v*ds(4)+g_n6 *v*ds(6)+g_n7 *v*ds(7)

# Compute solution
    u = Function(V)
    solve(a == L, u, bcs)
#solve(A, u.vector(), b)

    u_nodal_values = u.vector()
    u_array = u_nodal_values.array()
        
        #plot(u, interactive=True)
    DW=np.zeros((n))
    e1=np.zeros(n, dtype=np.float64)
    e2=np.zeros(n, dtype=np.float64)

    e1=u_array[G1]
    e2=u_array[G2]

    for i in range(n):
        e1[i]=e1[i]-phi
        e2[i]=e2[i]-phi

# Check if salt water has reached wells 1 and 2
    if np.min(e1)<DOLFIN_EPS:
        DW[0]=1
    if np.min(e2)<DOLFIN_EPS:
        DW[1]=1
    # Plot saltwater wave
    if flagplot==1:
        bmesh = BoundaryMesh(mesh, "exterior")
        bcoor = bmesh.coordinates()
        n_bpoints=len(bcoor)
        xb=[0.0 for i in range(n_bpoints)]
        yb=[0.0 for i in range(n_bpoints)]
        x_wave=[0.0 for i in range(n_bpoints)]
        y_wave=[0.0 for i in range(n_bpoints)]

        for i in range(n_bpoints):
            xb[i]=bcoor[i][0]
            yb[i]=bcoor[i][1]

        print 'Found ',n_bpoints,' Boundary points'
    #  Evaluate boundary cells
        print 'Found ',mesh.num_cells(),' Finite elemets'

        boundary_adjacent_cells = [myCell for myCell in cells(mesh) if any([myFacet.exterior() for myFacet in facets(myCell)])]

        k=3*len(boundary_adjacent_cells)
        xbv=[0 for i in range(k)]
        ybv=[0 for i in range(k)]
        ubv=[0 for i in range(k)]

        k=0
        for cell in boundary_adjacent_cells:
            for vertex in vertices(cell):
                ubv[k]=u_nodal_values[vertex.index()]
                xbv[k]=coordinates[vertex.index()][0]
                ybv[k]=coordinates[vertex.index()][1]
                k=k+1
        tvf=[0.0, 0.0, 0.0]
        flag=0
        k=0
        for cell in boundary_adjacent_cells:
            k=k+1
            tv=[-100.0, -100.0, -100.0]
            for vertex in vertices(cell):
                if tv[0]==-100.0:
                    tv[0]=u_nodal_values[vertex.index()]
                    x1br=coordinates[vertex.index()][0]
                    y1br=coordinates[vertex.index()][1]
                elif tv[1]==-100.0:
                    tv[1]=u_nodal_values[vertex.index()]
                    x2br=coordinates[vertex.index()][0]
                    y2br=coordinates[vertex.index()][1]
                else:
                    tv[2]=u_nodal_values[vertex.index()]
                    x3br=coordinates[vertex.index()][0]
                    y3br=coordinates[vertex.index()][1]
            max_u=np.max(tv)
            min_u=np.min(tv)

            flag_point=[0, 0, 0]
            if max_u>=phi and min_u<=phi:
                # This Cell is a possible  Wave Boundary Cell
                for i in range(n_bpoints):
                    if x1br==xb[i] and y1br==yb[i]:
                        flag_point[0]=1
                    elif x2br==xb[i] and y2br==yb[i]:
                        flag_point[1]=1
                    elif x3br==xb[i] and y3br==yb[i]:
                        flag_point[2]=1
                    if np.sum(flag_point)>1:
                        #print i,cell.index(),flag_point
                        break
                if np.sum(flag_point)>1:
                    for kk in range(3):
                        tvf[kk]=tv[kk]-phi
                    if flag_point[0]==1 and flag_point[1]==1:
                        v1=tvf[0]
                        v2=tvf[1]
                        fp=12
                    elif flag_point[0]==1 and flag_point[2]==1:
                        v2=tvf[2]
                        v1=tvf[0]
                        fp=13
                    else:
                        fp=23
                        v2=tvf[2]
                        v1=tvf[1]
                    if v1*v2<=0:
                        # The Wave Boundary Cell is found
                        cell_start=cell
                        # print "The Wave Boundary Cell is",cell.index(),flag_point,v1,v2
                        flag=1
                        break
            cell_prev=cell
            if flag==1:
                break
        if flag==0:
            #print "A Boundary Wave Cell can NOT be found"
            quit()
 
        if fp==23:
            temp=x1br
            x1br=x2br
            x2br=x3br
            x3br=temp
            temp=y1br
            y1br=y2br
            y2br=y3br
            y3br=temp
            temp=tv[0]
            tv[0]=tv[1]
            tv[1]=tv[2]
            tv[2]=temp
        elif fp==13:
            temp=x1br
            x1br=x3br
            x3br=temp
            temp=y1br
            y1br=y3br
            y3br=temp
            temp=tv[0]
            tv[0]=tv[2]
            tv[2]=temp

        p1=[x1br, y1br, tv[0]]
        p2=[x2br, y2br, tv[1]]
        p3=[x3br, y3br, tv[2]]
        l_line=(phi-tv[0])/(tv[1]-tv[0])
        xwave=x1br+l_line*(x2br-x1br)
        ywave= y1br+l_line*(y2br-y1br)
        p=[xwave,ywave, phi]
        pnew=[0.0,0.0]
        k=0
        x_wave[0]=xwave
        y_wave[0]=ywave
        flag_wave=0
        num_wave_points=1
        # Iterative procedure for Wave location
        while flag_wave==0:
            pnew, fp = Wave(p1,p2,p3,p)
            k=k+1
            x_wave[k]=pnew[0]
            y_wave[k]=pnew[1]
            plt.ion()
            plt.axis('scaled')
            plt.plot(WX[0],WY[0],'bs',WX[1],WY[1],'gs',WX[2],WY[2],'rs',WX[3],WY[3],'cs',xbv,ybv,'g.',xb, yb,'r.',x1br,y1br,'cs',x2br,y2br,'rs',x3br,y3br,'gs',x_wave,y_wave,'b.',GW1X[0],GW1Y[0],'r.',GW1X[1],GW1Y[1],'r.',GW1X[2],GW1Y[2],'r.',GW1X[3],GW1Y[3],'r.',GW2X[0],GW2Y[0],'r.',GW2X[1],GW2Y[1],'r.',GW2X[2],GW2Y[2],'r.',GW2X[3],GW2Y[3],'r.')
            plt.axis([-50, 6000, -50, 5000])
            plt.draw()
            plt.pause(0.0001)
            plt.clf()
            if fp==23:
                p1=p2
                p2=p3
            else:
                p2=p3
            p=[pnew[0],pnew[1],phi]

            cell_neighbors = sum((filter(lambda ci: ci != cell_start.index(),facet.entities(mesh.topology().dim())) for facet in facets(cell_start)),[])

            num_cell_neighbors=len(cell_neighbors)

            flag_new_cell=0
            for idc in range(num_cell_neighbors):
                for cell in cells(mesh):
                    if cell.index()==cell_neighbors[idc]:
                        c1=cell
                        break
        #        print 'The new Wave Cell is possible to be ',c1.index()
                p3i=[0,0,0]
                p3x=[0.0,0.0,0.0]
                p3y=[0.0,0.0,0.0]
                p3u=[0.0,0.0,0.0]
                id=0
                for vertex in vertices(c1):
                    p3x[id]=coordinates[vertex.index()][0]
                    p3y[id]=coordinates[vertex.index()][1]
                    p3u[id]=u_nodal_values[vertex.index()]
                    if (p3x[id]==p1[0] and p3y[id]==p1[1]) or (p3x[id]==p2[0] and p3y[id]==p2[1]):
                        p3i[id]=1
                    id=id+1
                
                if sum(p3i)==2:
                    cell_start=c1
                    for id in range(3):
                        if p3i[id]==0:
                            p3=[p3x[id],p3y[id],p3u[id]]
                            flag_new_cell=1
                if flag_new_cell==1:
                    break
            
            for cell in boundary_adjacent_cells:
                if cell==cell_start:
                    #  print 'Discovered the Wave Boundary Cell ',cell.index()
                    if num_wave_points>10:
                        flag_wave=1
            num_wave_points=num_wave_points+1

        pnew, fp = Wave(p1,p2,p3,p)
        k=k+1
        x_wave[k]=pnew[0]
        y_wave[k]=pnew[1]
        k=k+1
# print 'Found', k, ' wave points'
        if iflag==1:
            x_wave_value=[0.0 for i in range(k)]
            y_wave_value=[0.0 for i in range(k)]
            xn_wave_value=[0.0 for i in range(k)]
            yn_wave_value=[0.0 for i in range(k)]
            for i in range(k):
                x_wave_value[i]=x_wave[i]
                y_wave_value[i]=y_wave[i]

            np.savetxt('xw.txt', x_wave_value)
            np.savetxt('yw.txt', y_wave_value)
            fx=open('xw.txt',"r")
            xn_wave_value=fx.readlines()
            fx.close()
            fy=open('yw.txt',"r")
            yn_wave_value=fy.readlines()
            fy.close()

            plt.ioff()
            plt.axis('scaled')
            plt.plot(WX[0],WY[0],'bs',WX[1],WY[1],'gs',WX[2],WY[2],'rs',WX[3],WY[3],'cs',xbv,ybv,'g.',xb, yb,'r.',x1br,y1br,'cs',x2br,y2br,'rs',x3br,y3br,'gs',x_wave_value,y_wave_value,'b.-',xn_wave_value,yn_wave_value,'c.-',GW1X[0],GW1Y[0],'r.',GW1X[1],GW1Y[1],'r.',GW1X[2],GW1Y[2],'r.',GW1X[3],GW1Y[3],'r.',GW2X[0],GW2Y[0],'r.',GW2X[1],GW2Y[1],'r.',GW2X[2],GW2Y[2],'r.',GW2X[3],GW2Y[3],'r.')
            plt.axis([-50, 6000, -50, 5000])
            plt.show()

    return DW

