# This code has been developed under THALIS: MATENVMED project for Action 4.3 and co-financed by Greece and EU - NSRF 2007-2013.
# This FEniCS python code is used for solving the flow potential problem for Hersonisos aquifer problem with four wells using ALOPEX optimization scheme and is compatible for FEniCS software ver. 1.6.0 +.

from dolfin import *
import sys
sys.path.append('./')
import scipy as sc
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import random
from FPDE import *

# Define ALOPEX iterations
Alopex_steps=300
iterations=Alopex_steps+1

# Number of wells
n=4

flagplot=0 # Plot flag = 1 -> view the wave plot at every Alopex step

#Define hydraulic conductivity (K values) for each subdomain
k_values=[5.2,0.15,0.15,430.0,0.6048,0.15,5.2,172.8,430.0,8.64,12.96,12.96,0.15,0.432]

mesh = Mesh("H30.xml")
# Initialize mesh function for boundary domains
boundaries = MeshFunction('size_t', mesh,"H30_facet_region.xml")
subdomains = MeshFunction('size_t', mesh, "H30_physical_region.xml")
parameters["reorder_dofs_serial"] = False

coordinates = mesh.coordinates()

# Define Well and Guard points for each Well

G1=[0 for i in range(n)]
G2=[0 for i in range(n)]

DW=[0 in range(n)]
WX=[0 for i in range(n)]
WY=[0 for i in range(n)]
GW1X=[0 for i in range(n)]
GW1Y=[0 for i in range(n)]
GW2X=[0 for i in range(n)]
GW2Y=[0 for i in range(n)]



for vertex in vertices(mesh):
    if abs(coordinates[vertex.index()][0]-2141.852353482341)<0.000001 and abs(coordinates[vertex.index()][1]-2330.385219383018)<0.000001:
        print vertex.index(),"W1 node"
        WX[0]=coordinates[vertex.index()][0]
        WY[0]=coordinates[vertex.index()][1]
    if abs(coordinates[vertex.index()][0]-2995.724272522548)<0.000001 and abs(coordinates[vertex.index()][1]-1977.53353599431)<0.001:
        print vertex.index(),"W2 node"
        WX[1]=coordinates[vertex.index()][0]
        WY[1]=coordinates[vertex.index()][1]
    if abs(coordinates[vertex.index()][0]-2667.855171268895)<0.000001 and abs(coordinates[vertex.index()][1]-1758.188059685744)<0.000001:
        print vertex.index(),"W3 node"
        WX[2]=coordinates[vertex.index()][0]
        WY[2]=coordinates[vertex.index()][1]
    if abs(coordinates[vertex.index()][0]-2445.413536372673)<0.000001 and abs(coordinates[vertex.index()][1]-1258.952700816654)<0.000001:
        print vertex.index(),"W4 node"
        WX[3]=coordinates[vertex.index()][0]
        WY[3]=coordinates[vertex.index()][1]

    if abs(coordinates[vertex.index()][0]-2323.48370118305)<0.000001 and abs(coordinates[vertex.index()][1]-2327.413826616946)<0.000001:
        print vertex.index(),"GW1 1 node"
        G1[0]=vertex.index()
        GW1X[0]=coordinates[vertex.index()][0]
        GW1Y[0]=coordinates[vertex.index()][1]
    if abs(coordinates[vertex.index()][0]-2276.643841994124)<0.000001 and abs(coordinates[vertex.index()][1]-2443.776090655123)<0.000001:
        print vertex.index(),"GW1 2 node"
        G1[1]=vertex.index()
        GW1X[1]=coordinates[vertex.index()][0]
        GW1Y[1]=coordinates[vertex.index()][1]
    if abs(coordinates[vertex.index()][0]-2142.887333565796 )<0.000001 and abs(coordinates[vertex.index()][1]-2518.065429194561)<0.000001:
        print vertex.index(),"GW1 3 node"
        G1[2]=vertex.index()
        GW1X[2]=coordinates[vertex.index()][0]
        GW1Y[2]=coordinates[vertex.index()][1]
    if abs(coordinates[vertex.index()][0]-2031.688777579124)<0.000001 and abs(coordinates[vertex.index()][1]-2479.166618173772)<0.000001:
        print vertex.index(),"GW1 4 node"
        G1[3]=vertex.index()
        GW1X[3]=coordinates[vertex.index()][0]
        GW1Y[3]=coordinates[vertex.index()][1]

    if abs(coordinates[vertex.index()][0]-3152.859512200966)<0.000001 and abs(coordinates[vertex.index()][1]-1911.423708309243)<0.000001:
        print vertex.index(),"GW2 1 node"
        G2[0]=vertex.index()
        GW2X[0]=coordinates[vertex.index()][0]
        GW2Y[0]=coordinates[vertex.index()][1]
    if abs(coordinates[vertex.index()][0]-3178.883875216069)<0.000001 and abs(coordinates[vertex.index()][1]-1995.914433679679)<0.000001:
        print vertex.index(),"GW2 2 node"
        G2[1]=vertex.index()
        GW2X[1]=coordinates[vertex.index()][0]
        GW2Y[1]=coordinates[vertex.index()][1]
    if abs(coordinates[vertex.index()][0]-3147.904426900827)<0.000001 and abs(coordinates[vertex.index()][1]-2096.499619546321)<0.000001:
        print vertex.index(),"GW2 3 node"
        G2[2]=vertex.index()
        GW2X[2]=coordinates[vertex.index()][0]
        GW2Y[2]=coordinates[vertex.index()][1]
    if abs(coordinates[vertex.index()][0]-2980.435866323473)<0.000001 and abs(coordinates[vertex.index()][1]-2183.535929936874)<0.000001:
        print vertex.index(),"GW2 4 node"
        G2[3]=vertex.index()
        GW2X[3]=coordinates[vertex.index()][0]
        GW2Y[3]=coordinates[vertex.index()][1]


WellPoint=np.zeros((n,n))
for i in range(n):
    WellPoint[i][0]=WX[i]
    WellPoint[i][1]=WY[i]

# Set finite element length
sh=30.0
# Vector KK contains k values for subdomains
KK=np.zeros(len(subdomains.array()))
help = np.asarray(subdomains.array(), dtype=np.int32)
KK[:]=np.choose(help, k_values)

# Define total recharge capability (m^3/day)
Qtotal=10000.0

# Define min and max volume pumped out per day for each well
Qlocalmin=[756.0,172.8, 756.0, 43.8]
Qlocalmax=[2520.0, 756.0, 2520.0, 146.0]

# Define penalties management
PenQlocalMax=0.95
PenQlocalMin=1.05
PenQlocal=1.05
PenXmov=0.95
PenSecXmov=0.95
# Preallocated variables
yit=np.zeros(iterations)
#Dist=np.zeros((n))
#xtgood=np.zeros((301))
Q=np.zeros((n,iterations))
Profit=np.zeros((iterations))
Q_opt=np.zeros((n))
Q_k=np.zeros((n))
ProfitOpt=0
kOpt=1
feedbackFactor=np.zeros((n,iterations))
noiseFactor=np.zeros((n,iterations))
# Preallocated flag variables
flagQlocalmin=np.zeros((n,iterations))
flagP1=np.zeros((n,iterations))
flagP2=np.zeros((n,iterations))
flagQlocalmax=np.zeros((n,iterations))
flagQtotal=np.zeros((n,iterations))
# Define acceleration factor
c=1.0
# Define initial pumping rates
for i in range(n):
    Q[i][0]=300.0+0.01*Qlocalmax[i]*(-1.0+2.0*random.uniform(0.1, 1.0))
    Q[i][1]=320.0+0.01*Qlocalmax[i]*np.fabs(-1.0+2.0*random.uniform(0.1, 1.0))

# ALOPEX Initial Iterations
k = 1
while k < 3:
    sm1=0.0
    sm=0.0
    for i in range(n):
        sm=sm+Q[i][k-1]
        sm1=sm1+Qlocalmax[i]
    Profit[k-1]=exp(-pow((sm1-sm)/sm1,2))
    k=k+1

Pr_opt=Profit[k-1]
for i in range(n):
    Q_opt[i]=Q[i][k-1]

# ALOPEX   Iterations
k = 3
kopt=3

while k < iterations :
    for i in range(0, n):
        c=1.0
        if Profit[k-2]-Profit[k-3] < 0.0:
            c=-c
        feedbackFactor[i][k-1]=c*(Q[i][k-2]-Q[i][k-3])
        noiseFactor[i][k-1]=0.015*Q[i][k-2]*(-0.5+1.5*random.uniform(0.1, 1.0))
        Q[i][k-1]=Q[i][k-2] + feedbackFactor[i][k-1]+noiseFactor[i][k-1]
    
    for i in range(0, n):
        if Q[i][k-1] > Qlocalmax[i]:
            Q[i][k-1]=Qlocalmax[i]*PenQlocalMax
            flagQlocalmax[i][k-1]=1
        if Q[i][k-1] < Qlocalmin[i]:
            Q[i][k-1]=Qlocalmin[i]*PenQlocalMin
            flagQlocalmin[i][k-1]=1
    Skm1=0.0
    for i in range(n):
        Skm1=Skm1+Q[i][k-2]
   
    for i in range(0, n):
        S=Skm1-Q[i][k-2]+Q[i][k-1]
        if S>Qtotal:
            Q[i][k-1]=Q[i][k-1]-PenQlocal*(S-Qtotal)
            flagQtotal[i][k-1]=1
        Skp1=S


    sq=0.0
    for i in range(n):
        sq=sq+Q[i][k-1]
        Q_k[i]=Q[i][k-1]
    
#    print 'ALOPEX Iteration', k,' total Q =',sq,' m3/day'
#    print Q_k
# Run Fenics module

    DW =FPDE(Q_k,KK,sh,flagplot,WX,WY,GW1X,GW1Y,GW2X,GW2Y,G1,G2,0)

    flag2=0
    for i in range(n):
        if DW[i]==1:
            Q[i][k-1]=Q[i][k-1]*PenXmov
            flagP1[i][k-1]=1
            flag2=1

    if flag2==1:
        flag2=0
        sq=0.0
        for i in range(n):
            Q_k[i]=Q[i][k-1]
            sq=sq+Q_k[i]
        print 'ALOPEX Iteration', k,' after Penalty 1 with total Q =',sq,' m3/day'
        print Q_k

        DW =FPDE(Q_k,KK,sh,flagplot,WX,WY,GW1X,GW1Y,GW2X,GW2Y,G1,G2,0)

        flag3=0
        iflood=0
        kflood=0
        for i in range(n):
            if DW[i]==1:
                flag3=1
        
        if flag3==1:
            flag3=0
            for i in range(n):
                Q[i][k-1]=Q[i][k-1]*PenSecXmov
                flagP2[i][k-1]=1
            print '------   Penalty 2 has been already applied   ------'

    sm=0.0
    sm1=0.0
    for i in range(n):
        sm=sm+Q[i][k-1]
        sm1=sm1+Qlocalmax[i]
    Profit[k-1]=exp(-pow((sm1-sm)/sm1,2))
    flagNopt=0
    for i in range(n):
        if Q[i][k-1] < Qlocalmin[i] or Q[i][k-1] > Qlocalmax[i] or DW[i]==1:
            flagNopt=1


    if Profit[k-1]>Pr_opt and flagNopt==0:
        Pr_opt=Profit[k-1]
        kopt=k
        for i in range(n):
            Q_opt[i]=Q[i][k-1]

    k=k+1
print 'Optimal Q values for Well1 (',WellPoint[0][0],WellPoint[0][1],')=',Q_opt[0],'m3/day'
print 'Optimal Q values for Well2 (',WellPoint[1][0],WellPoint[1][1],')=',Q_opt[1],'m3/day'
print 'Optimal Q values for Well3 (',WellPoint[2][0],WellPoint[2][1],')=',Q_opt[2],'m3/day'
print 'Optimal Q values for Well4 (',WellPoint[3][0],WellPoint[3][1],')=',Q_opt[3],'m3/day'

OptQ=0.0
for i in range(n):
    OptQ=OptQ+Q_opt[i]
print 'Optimal Total Q values =',OptQ,' m3/day'
print 'Optimal ALOPEX Step =',kopt
flagplot=1
DW =FPDE(Q_k,KK,sh,1,WX,WY,GW1X,GW1Y,GW2X,GW2Y,G1,G2,1)
for i in range(iterations):
    yit[i]=i

plt.plot(yit[1:iterations],feedbackFactor[1][0:iterations-1], 'm-',label='Feedback Factor')
plt.plot(yit[1:iterations],noiseFactor[1][0:iterations-1], 'k-',label='NoiseFactor')
plt.title('F-N for Wells')
plt.legend(loc='best', numpoints=1)
plt.show()


plt.plot( yit[1:iterations],Profit[0:iterations-1], 'g.-',label='Profit function')
plt.title('Profit function')
plt.legend(loc='best', numpoints=1)
plt.show()

plt.plot( yit[1:iterations],Q[0][0:iterations-1], 'r.-',label='Q for Well1')
plt.title('Q for Well1')
plt.legend(loc='best', numpoints=1)
plt.show()

plt.plot( yit[1:iterations],Q[1][0:iterations-1], 'b.-',label='Q for Well2')
plt.title('Q for Well2')
plt.legend(loc='best', numpoints=1)
plt.show()

plt.plot( yit[1:iterations],Q[2][0:iterations-1], 'g.-',label='Q for Well3')
plt.title('Q for Well3')
plt.legend(loc='best', numpoints=1)
plt.show()
plt.plot( yit[1:iterations],Q[3][0:iterations-1], 'r.-',label='Q for Well4')
plt.title('Q for Well4')
plt.legend(loc='best', numpoints=1)
plt.show()

plt.plot( yit[1:iterations],Q[0][0:iterations-1], 'r-',label='Well1')
plt.plot(yit[1:iterations],Q[1][0:iterations-1], 'b--',label='Well2')
plt.plot(yit[1:iterations],Q[2][0:iterations-1], 'g-',label='Well3')
plt.plot(yit[1:iterations],Q[3][0:iterations-1], 'm--',label='Well4')
plt.title('Q for Wells')
plt.legend(loc='best', numpoints=1)
plt.show()

