# This code has been developed under THALIS: MATENVMED project for Action 4.3 and co-financed by Greece and EU - NSRF 2007-2013.
# This FEniCS python code is used for solving the flow potential problem for Kalymnos aquifer problem with five wells using ALOPEX optimization scheme and is compatible for FEniCS software ver. 1.6.0 +.

from dolfin import *
import sys
sys.path.append('./')
import scipy as sc
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import random
from FPDE import *

#Define  hydraulic conductivity (K values) for each subdomain
k_values = [25.0, 35.0, 50.0, 75.0, 50.0]
#k_values = [100.0, 100.0, 100.0, 100.0, 100.0]

# Define ALOPEX iterations
Alopex_steps=50
iterations=Alopex_steps+1

# Number of wells
n=5

ierror=0 # 1 for error calculation for uniform K
flagplot=1 # Plot flag = 1 -> view the wave plot at every Alopex step
iflag=1 # 1 -> calculate the wave at every Alopex step

# Define Subdomains of K for Fenics
mesh = RectangleMesh(Point(0.0, 0.0),Point(7000.0, 3000.0), 700, 300, "right")
parameters["reorder_dofs_serial"] = False
    
# Initialize mesh function for interior domains
domains = CellFunction("size_t", mesh)
domains.set_all(0)

subdomains = MeshFunction('size_t', mesh, 2)

# Define Subdomains
class Omega0(SubDomain):
    def inside(self, x, on_boundary):
        return True if x[1] >= 1950 else False

class Omega1(SubDomain):
    def inside(self, x, on_boundary):
        return True if x[1] >= 1200 and x[1] <= 1950 and x[0] <= 6000 else False

class Omega2(SubDomain):
    def inside(self, x, on_boundary):
        return True if x[1] <= 1200 and x[0] >= 2625 else False

class Omega3(SubDomain):
    def inside(self, x, on_boundary):
        return True if x[1] <= 1200  and x[0] <= 2625 else False

class Omega4(SubDomain):
    def inside(self, x, on_boundary):
        return True if x[1] >= 1200  and x[1] <= 1950 and x[0] >= 6000 else False

subdomain0 = Omega0()
subdomain0.mark(subdomains, 0)
subdomain1 = Omega1()
subdomain1.mark(subdomains, 1)
subdomain2 = Omega2()
subdomain2.mark(subdomains, 2)
subdomain3 = Omega3()
subdomain3.mark(subdomains, 3)
subdomain4 = Omega4()
subdomain4.mark(subdomains, 4)

# Vector KK contains k values for subdomains
KK=np.zeros(len(subdomains.array()))
help = np.asarray(subdomains.array(), dtype=np.int32)
KK[:]=np.choose(help, k_values)

# Define well-points
WellPoint=[[0 for i in range(n)] for j in range(n)]
WellPoint[0][0]=2640.0
WellPoint[0][1]=1560.0
WellPoint[1][0]=3340.0
WellPoint[1][1]=2200.0
WellPoint[2][0]=3920.0
WellPoint[2][1]=960.0
WellPoint[3][0]=4620.0
WellPoint[3][1]=2460.0
WellPoint[4][0]=4860.0
WellPoint[4][1]=1580.0

# Define critical distance from well-points
CriticalWell=[0 for i in range(n)]
CriticalWell[0]=400.0
CriticalWell[1]=400.0
CriticalWell[2]=400.0
CriticalWell[3]=400.0
CriticalWell[4]=400.0

# Define Guard points for each Well
G1x=[0.0 for i in range(11)]
G1y=[0.0 for i in range(11)]
G1x[0]=WellPoint[0][0]-CriticalWell[0]
G1y[0]=WellPoint[0][1]
G1x[1]=WellPoint[0][0]
G1y[1]=WellPoint[0][1]-CriticalWell[0]
G1x[2]=WellPoint[0][0]
G1y[2]=WellPoint[0][1]+CriticalWell[0]
G1x[3]=WellPoint[0][0]-0.5*CriticalWell[0]
G1y[3]=1220.0
G1x[4]=WellPoint[0][0]-0.5*CriticalWell[0]
G1y[4]=1900.0
G1x[5]=WellPoint[0][0]+0.5*CriticalWell[0]
G1y[5]=1220.0
G1x[6]=WellPoint[0][0]+0.5*CriticalWell[0]
G1y[6]=1900.0
G1x[7]=WellPoint[0][0]-340.0
G1y[7]=1340.0
G1x[8]=WellPoint[0][0]-340.0
G1y[8]=1780.0
G1x[9]=WellPoint[0][0]-380.0
G1y[9]=1680.0
G1x[10]=WellPoint[0][0]-380.0
G1y[10]=1440.0

G2x=[0.0 for i in range(11)]
G2y=[0.0 for i in range(11)]
G2x[0]=WellPoint[1][0]-CriticalWell[1]
G2y[0]=WellPoint[1][1]
G2x[1]=WellPoint[1][0]
G2y[1]=WellPoint[1][1]-CriticalWell[1]
G2x[2]=WellPoint[1][0]
G2y[2]=WellPoint[1][1]+CriticalWell[1]
G2x[3]=WellPoint[1][0]-0.5*CriticalWell[1]
G2y[3]=1220.0+WellPoint[1][1]-WellPoint[0][1]
G2x[4]=WellPoint[1][0]-0.5*CriticalWell[1]
G2y[4]=1900.0+WellPoint[1][1]-WellPoint[0][1]
G2x[5]=WellPoint[1][0]+0.5*CriticalWell[1]
G2y[5]=1220.0+WellPoint[1][1]-WellPoint[0][1]
G2x[6]=WellPoint[1][0]+0.5*CriticalWell[1]
G2y[6]=1900.0+WellPoint[1][1]-WellPoint[0][1]
G2x[7]=WellPoint[1][0]-340.0
G2y[7]=1340.0+WellPoint[1][1]-WellPoint[0][1]
G2x[8]=WellPoint[1][0]-340.0
G2y[8]=1760.0+WellPoint[1][1]-WellPoint[0][1]
G2x[9]=WellPoint[1][0]+360.0
G2y[9]=1760.0+WellPoint[1][1]-WellPoint[0][1]
G2x[10]=WellPoint[1][0]+360.0
G2y[10]=1380.0+WellPoint[1][1]-WellPoint[0][1]

G3x=[0.0 for i in range(11)]
G3y=[0.0 for i in range(11)]
G3x[0]=WellPoint[2][0]-CriticalWell[2]
G3y[0]=WellPoint[2][1]
G3x[1]=WellPoint[2][0]
G3y[1]=WellPoint[2][1]-CriticalWell[2]
G3x[2]=WellPoint[2][0]
G3y[2]=WellPoint[2][1]+CriticalWell[2]
G3x[3]=WellPoint[2][0]-0.5*CriticalWell[2]
G3y[3]=1220.0+WellPoint[2][1]-WellPoint[0][1]
G3x[4]=WellPoint[2][0]-0.5*CriticalWell[2]
G3y[4]=1900.0+WellPoint[2][1]-WellPoint[0][1]
G3x[5]=WellPoint[2][0]+0.5*CriticalWell[2]
G3y[5]=1220.0+WellPoint[2][1]-WellPoint[0][1]
G3x[6]=WellPoint[2][0]+0.5*CriticalWell[2]
G3y[6]=1900.0+WellPoint[2][1]-WellPoint[0][1]
G3x[7]=WellPoint[2][0]-340.0
G3y[7]=1340.0+WellPoint[2][1]-WellPoint[0][1]
G3x[8]=WellPoint[2][0]-340.0
G3y[8]=1760.0+WellPoint[2][1]-WellPoint[0][1]
G3x[9]=WellPoint[2][0]+360.0
G3y[9]=1760.0+WellPoint[2][1]-WellPoint[0][1]
G3x[10]=WellPoint[2][0]+360.0
G3y[10]=1380.0+WellPoint[2][1]-WellPoint[0][1]

G4x=[0.0 for i in range(11)]
G4y=[0.0 for i in range(11)]
G4x[0]=WellPoint[3][0]-CriticalWell[3]
G4y[0]=WellPoint[3][1]
G4x[1]=WellPoint[3][0]
G4y[1]=WellPoint[3][1]-CriticalWell[3]
G4x[2]=WellPoint[3][0]
G4y[2]=WellPoint[3][1]+CriticalWell[3]
G4x[3]=WellPoint[3][0]-0.5*CriticalWell[3]
G4y[3]=1220.0+WellPoint[3][1]-WellPoint[0][1]
G4x[4]=WellPoint[3][0]-0.5*CriticalWell[3]
G4y[4]=1900.0+WellPoint[3][1]-WellPoint[0][1]
G4x[5]=WellPoint[3][0]+0.5*CriticalWell[3]
G4y[5]=1220.0+WellPoint[3][1]-WellPoint[0][1]
G4x[6]=WellPoint[3][0]+0.5*CriticalWell[3]
G4y[6]=1900.0+WellPoint[3][1]-WellPoint[0][1]
G4x[7]=WellPoint[3][0]-340.0
G4y[7]=1340.0+WellPoint[3][1]-WellPoint[0][1]
G4x[8]=WellPoint[3][0]-340.0
G4y[8]=1760.0+WellPoint[3][1]-WellPoint[0][1]
G4x[9]=WellPoint[3][0]+360.0
G4y[9]=1760.0+WellPoint[3][1]-WellPoint[0][1]
G4x[10]=WellPoint[3][0]+360.0
G4y[10]=1380.0+WellPoint[3][1]-WellPoint[0][1]

G5x=[0.0 for i in range(11)]
G5y=[0.0 for i in range(11)]
G5x[0]=WellPoint[4][0]-CriticalWell[4]
G5y[0]=WellPoint[4][1]
G5x[1]=WellPoint[4][0]
G5y[1]=WellPoint[4][1]-CriticalWell[4]
G5x[2]=WellPoint[4][0]
G5y[2]=WellPoint[4][1]+CriticalWell[4]
G5x[3]=WellPoint[4][0]-0.5*CriticalWell[4]
G5y[3]=1220.0+WellPoint[4][1]-WellPoint[0][1]
G5x[4]=WellPoint[4][0]-0.5*CriticalWell[4]
G5y[4]=1900.0+WellPoint[4][1]-WellPoint[0][1]
G5x[5]=WellPoint[4][0]+0.5*CriticalWell[4]
G5y[5]=1220.0+WellPoint[4][1]-WellPoint[0][1]
G5x[6]=WellPoint[4][0]+0.5*CriticalWell[4]
G5y[6]=1900.0+WellPoint[4][1]-WellPoint[0][1]
G5x[7]=WellPoint[4][0]-340.0
G5y[7]=1340.0+WellPoint[4][1]-WellPoint[0][1]
G5x[8]=WellPoint[4][0]-340.0
G5y[8]=1760.0+WellPoint[4][1]-WellPoint[0][1]
G5x[9]=WellPoint[4][0]+360.0
G5y[9]=1760.0+WellPoint[4][1]-WellPoint[0][1]
G5x[10]=WellPoint[4][0]+360.0
G5y[10]=1380.0+WellPoint[4][1]-WellPoint[0][1]

G1=[0 for i in range(11)]
G2=[0 for i in range(11)]
G3=[0 for i in range(11)]
G4=[0 for i in range(11)]
G5=[0 for i in range(11)]

coordinates = mesh.coordinates()

for v in vertices(mesh):
    if coordinates[v.index()][0]==G1x[0] and coordinates[v.index()][1]==G1y[0]:
        G1[0]=v.index()
    if coordinates[v.index()][0]==G1x[1] and coordinates[v.index()][1]==G1y[1]:
        G1[1]=v.index()
    if coordinates[v.index()][0]==G1x[2] and coordinates[v.index()][1]==G1y[2]:
        G1[2]=v.index()
    if coordinates[v.index()][0]==G1x[3] and coordinates[v.index()][1]==G1y[3]:
        G1[3]=v.index()
    if coordinates[v.index()][0]==G1x[4] and coordinates[v.index()][1]==G1y[4]:
        G1[4]=v.index()
    if coordinates[v.index()][0]==G1x[5] and coordinates[v.index()][1]==G1y[5]:
        G1[5]=v.index()
    if coordinates[v.index()][0]==G1x[6] and coordinates[v.index()][1]==G1y[6]:
        G1[6]=v.index()
    if coordinates[v.index()][0]==G1x[7] and coordinates[v.index()][1]==G1y[7]:
        G1[7]=v.index()
    if coordinates[v.index()][0]==G1x[8] and coordinates[v.index()][1]==G1y[8]:
        G1[8]=v.index()
    if coordinates[v.index()][0]==G1x[9] and coordinates[v.index()][1]==G1y[9]:
        G1[9]=v.index()
    if coordinates[v.index()][0]==G1x[10] and coordinates[v.index()][1]==G1y[10]:
        G1[10]=v.index()

    if coordinates[v.index()][0]==G2x[0] and coordinates[v.index()][1]==G2y[0]:
        G2[0]=v.index()
    if coordinates[v.index()][0]==G2x[1] and coordinates[v.index()][1]==G2y[1]:
        G2[1]=v.index()
    if coordinates[v.index()][0]==G2x[2] and coordinates[v.index()][1]==G2y[2]:
        G2[2]=v.index()
    if coordinates[v.index()][0]==G2x[3] and coordinates[v.index()][1]==G2y[3]:
        G2[3]=v.index()
    if coordinates[v.index()][0]==G2x[4] and coordinates[v.index()][1]==G2y[4]:
        G2[4]=v.index()
    if coordinates[v.index()][0]==G2x[5] and coordinates[v.index()][1]==G2y[5]:
        G2[5]=v.index()
    if coordinates[v.index()][0]==G2x[6] and coordinates[v.index()][1]==G2y[6]:
        G2[6]=v.index()
    if coordinates[v.index()][0]==G2x[7] and coordinates[v.index()][1]==G2y[7]:
        G2[7]=v.index()
    if coordinates[v.index()][0]==G2x[8] and coordinates[v.index()][1]==G2y[8]:
        G2[8]=v.index()
    if coordinates[v.index()][0]==G2x[9] and coordinates[v.index()][1]==G2y[9]:
        G2[9]=v.index()
    if coordinates[v.index()][0]==G2x[10] and coordinates[v.index()][1]==G2y[10]:
        G2[10]=v.index()

    if coordinates[v.index()][0]==G3x[0] and coordinates[v.index()][1]==G3y[0]:
        G3[0]=v.index()
    if coordinates[v.index()][0]==G3x[1] and coordinates[v.index()][1]==G3y[1]:
        G3[1]=v.index()
    if coordinates[v.index()][0]==G3x[2] and coordinates[v.index()][1]==G3y[2]:
        G3[2]=v.index()
    if coordinates[v.index()][0]==G3x[3] and coordinates[v.index()][1]==G3y[3]:
        G3[3]=v.index()
    if coordinates[v.index()][0]==G3x[4] and coordinates[v.index()][1]==G3y[4]:
        G3[4]=v.index()
    if coordinates[v.index()][0]==G3x[5] and coordinates[v.index()][1]==G3y[5]:
        G3[5]=v.index()
    if coordinates[v.index()][0]==G3x[6] and coordinates[v.index()][1]==G3y[6]:
        G3[6]=v.index()
    if coordinates[v.index()][0]==G3x[7] and coordinates[v.index()][1]==G3y[7]:
        G3[7]=v.index()
    if coordinates[v.index()][0]==G3x[8] and coordinates[v.index()][1]==G3y[8]:
        G3[8]=v.index()
    if coordinates[v.index()][0]==G3x[9] and coordinates[v.index()][1]==G3y[9]:
        G3[9]=v.index()
    if coordinates[v.index()][0]==G3x[10] and coordinates[v.index()][1]==G3y[10]:
        G3[10]=v.index()


    if coordinates[v.index()][0]==G4x[0] and coordinates[v.index()][1]==G4y[0]:
        G4[0]=v.index()
    if coordinates[v.index()][0]==G4x[1] and coordinates[v.index()][1]==G4y[1]:
        G4[1]=v.index()
    if coordinates[v.index()][0]==G4x[2] and coordinates[v.index()][1]==G4y[2]:
        G4[2]=v.index()
    if coordinates[v.index()][0]==G4x[3] and coordinates[v.index()][1]==G4y[3]:
        G4[3]=v.index()
    if coordinates[v.index()][0]==G4x[4] and coordinates[v.index()][1]==G4y[4]:
        G4[4]=v.index()
    if coordinates[v.index()][0]==G4x[5] and coordinates[v.index()][1]==G4y[5]:
        G4[5]=v.index()
    if coordinates[v.index()][0]==G4x[6] and coordinates[v.index()][1]==G4y[6]:
        G4[6]=v.index()
    if coordinates[v.index()][0]==G4x[7] and coordinates[v.index()][1]==G4y[7]:
        G4[7]=v.index()
    if coordinates[v.index()][0]==G4x[8] and coordinates[v.index()][1]==G4y[8]:
        G4[8]=v.index()
    if coordinates[v.index()][0]==G4x[9] and coordinates[v.index()][1]==G4y[9]:
        G4[9]=v.index()
    if coordinates[v.index()][0]==G4x[10] and coordinates[v.index()][1]==G4y[10]:
        G4[10]=v.index()
    
    if coordinates[v.index()][0]==G5x[0] and coordinates[v.index()][1]==G5y[0]:
        G5[0]=v.index()
    if coordinates[v.index()][0]==G5x[1] and coordinates[v.index()][1]==G5y[1]:
        G5[1]=v.index()
    if coordinates[v.index()][0]==G5x[2] and coordinates[v.index()][1]==G5y[2]:
        G5[2]=v.index()
    if coordinates[v.index()][0]==G5x[3] and coordinates[v.index()][1]==G5y[3]:
        G5[3]=v.index()
    if coordinates[v.index()][0]==G5x[4] and coordinates[v.index()][1]==G5y[4]:
        G5[4]=v.index()
    if coordinates[v.index()][0]==G5x[5] and coordinates[v.index()][1]==G5y[5]:
        G5[5]=v.index()
    if coordinates[v.index()][0]==G5x[6] and coordinates[v.index()][1]==G5y[6]:
        G5[6]=v.index()
    if coordinates[v.index()][0]==G5x[7] and coordinates[v.index()][1]==G5y[7]:
        G5[7]=v.index()
    if coordinates[v.index()][0]==G5x[8] and coordinates[v.index()][1]==G5y[8]:
        G5[8]=v.index()
    if coordinates[v.index()][0]==G5x[9] and coordinates[v.index()][1]==G5y[9]:
        G5[9]=v.index()
    if coordinates[v.index()][0]==G5x[10] and coordinates[v.index()][1]==G5y[10]:
        G5[10]=v.index()
DW=[0 in range(n)]
# Define total recharge capability (m^3/day)
Qtotal=20000.0

# Define min and max volume pumped out per day for each well
Qlocalmin=[200.0 for i in range(n)]
Qlocalmax=[2500.0 for i in range(n)]

# Define penalties management
PenQlocalMax=0.95
PenQlocalMin=1.05
PenQlocal=1.05
PenXmov=0.95
PenSecXmov=0.95
# Preallocated variables
yit=np.zeros(iterations)
Dist=np.zeros((n))
xtgood=np.zeros((301))
Q=np.zeros((n,iterations))
Profit=np.zeros((iterations))
Q_opt=np.zeros((n))
Q_k=np.zeros((n))
ProfitOpt=0
kOpt=1
feedbackFactor=np.zeros((n,iterations))
noiseFactor=np.zeros((n,iterations))
# Preallocated flag variables
flagQlocalmin=np.zeros((n,iterations))
flagP1=np.zeros((n,iterations))
flagP2=np.zeros((n,iterations))
flagQlocalmax=np.zeros((n,iterations))
flagQtotal=np.zeros((n,iterations))
# Define acceleration factor
c=1.0

# Define initial pumping rates
for i in range(n):
    Q[i][0]=300.0+0.01*Qlocalmax[i]*(-1.0+2.0*random.uniform(0.1, 1.0))
    Q[i][1]=320.0+0.01*Qlocalmax[i]*np.fabs(-1.0+2.0*random.uniform(0.1, 1.0))

# ALOPEX Initial Iterations
k = 1
while k < 3:
    sm1=0.0
    sm=0.0
    for i in range(n):
        sm=sm+Q[i][k-1]
        sm1=sm1+Qlocalmax[i]
    Profit[k-1]=exp(-pow((sm1-sm)/sm1,2))
    k=k+1

Pr_opt=Profit[k-1]
for i in range(n):
    Q_opt[i]=Q[i][k-1]

# ALOPEX   Iterations
k = 3
kopt=3

while k < iterations :
    for i in range(0, n):
        c=1.0
        if Profit[k-2]-Profit[k-3] < 0.0:
            c=-c
        feedbackFactor[i][k-1]=c*(Q[i][k-2]-Q[i][k-3])
        noiseFactor[i][k-1]=0.015*Q[i][k-2]*(-0.5+1.5*random.uniform(0.1, 1.0))
        Q[i][k-1]=Q[i][k-2] + feedbackFactor[i][k-1]+noiseFactor[i][k-1]
    
    for i in range(0, n):
        if Q[i][k-1] > Qlocalmax[i]:
            Q[i][k-1]=Qlocalmax[i]*PenQlocalMax
            flagQlocalmax[i][k-1]=1
        if Q[i][k-1] < Qlocalmin[i]:
            Q[i][k-1]=Qlocalmin[i]*PenQlocalMin
            flagQlocalmin[i][k-1]=1
    Skm1=0.0
    for i in range(n):
        Skm1=Skm1+Q[i][k-2]
   
    for i in range(0, n):
        S=Skm1-Q[i][k-2]+Q[i][k-1]
        if S>Qtotal:
            Q[i][k-1]=Q[i][k-1]-PenQlocal*(S-Qtotal)
            flagQtotal[i][k-1]=1
        Skp1=S


    sq=0.0
    for i in range(n):
        sq=sq+Q[i][k-1]
        Q_k[i]=Q[i][k-1]
    
    print 'ALOPEX Iteration', k,' total Q =',sq,' m3/day'
    print Q_k
# Run Solver module

    xtgood, DW =FPDE(Q_k,k_values,flagplot,0,KK,G1x,G1y,G2x,G2y,G3x,G3y,G4x,G4y,G5x,G5y,G1,G2,G3,G4,G5,iflag,ierror)

    flag2=0
    for i in range(n):
        if DW[i]==1:
            Q[i][k-1]=Q[i][k-1]*PenXmov
            flagP1[i][k-1]=1
            flag2=1

    if flag2==1:
        flag2=0
        sq=0.0
        for i in range(n):
            Q_k[i]=Q[i][k-1]
            sq=sq+Q_k[i]
        print 'ALOPEX Iteration', k,' after Penalty 1 with total Q =',sq,' m3/day'
        print Q_k
# Run Solver module
        xtgood, DW =FPDE(Q_k,k_values,flagplot,1,KK,G1x,G1y,G2x,G2y,G3x,G3y,G4x,G4y,G5x,G5y,G1,G2,G3,G4,G5,iflag,0)

        flag3=0
        iflood=0
        kflood=0
        for i in range(n):
            if DW[i]==1:
                if iflood < WellPoint[i][0]:
                    iflood=WellPoint[i][0]
                    kflood=i
                flag3=1

        if flag3==1:
            flag3=0
            print '------   Penalty 2 has been already applied   ------'
            for i in range(n):
                if WellPoint[i][0] > iflood:
                    Q[i][k-1]=Q[i][k-1]*PenSecXmov
                    flagP2[i][k-1]=1

    sm=0.0
    sm1=0.0
    for i in range(n):
        sm=sm+Q[i][k-1]
        sm1=sm1+Qlocalmax[i]
    Profit[k-1]=exp(-pow((sm1-sm)/sm1,2))
    flagNopt=0
    for i in range(n):
        if Q[i][k-1] < Qlocalmin[i] or Q[i][k-1] > Qlocalmax[i] or DW[i]==1:
            flagNopt=1


    if Profit[k-1]>Pr_opt and flagNopt==0:
        Pr_opt=Profit[k-1]
        kopt=k
        for i in range(n):
            Q_opt[i]=Q[i][k-1]

    k=k+1
# End of Alopex iteration
print 'Optimal Q values for Well1 (',WellPoint[0][0],WellPoint[0][1],')=',Q_opt[0],'m3/day'
print 'Optimal Q values for Well2 (',WellPoint[1][0],WellPoint[1][1],')=',Q_opt[1],'m3/day'
print 'Optimal Q values for Well3 (',WellPoint[2][0],WellPoint[2][1],')=',Q_opt[2],'m3/day'
print 'Optimal Q values for Well4 (',WellPoint[3][0],WellPoint[3][1],')=',Q_opt[3],'m3/day'
print 'Optimal Q values for Well5 (',WellPoint[4][0],WellPoint[4][1],')=',Q_opt[4],'m3/day'
OptQ=0.0
for i in range(n):
    OptQ=OptQ+Q_opt[i]
print 'Optimal Total Q values =',OptQ,' m3/day'
print 'Optimal ALOPEX Step =',kopt
flagplot=1
# Run Solver module with plot option for optimal Q's
print Q_opt
xtgood, DW =FPDE(Q_opt,k_values,flagplot,2,KK,G1x,G1y,G2x,G2y,G3x,G3y,G4x,G4y,G5x,G5y,G1,G2,G3,G4,G5,1,0)

for i in range(iterations):
    yit[i]=i

plt.plot(yit[1:iterations],feedbackFactor[1][0:iterations-1], 'm-',label='Feedback Factor')
plt.plot(yit[1:iterations],noiseFactor[1][0:iterations-1], 'k-',label='NoiseFactor')
plt.title('F-N for Wells')
plt.legend(loc='best', numpoints=1)
plt.show()


plt.plot( yit[1:iterations],Profit[0:iterations-1], 'g.-',label='Profit function')
plt.title('Profit function')
plt.legend(loc='best', numpoints=1)
plt.show()

plt.plot( yit[1:iterations],Q[0][0:iterations-1], 'r.-',label='Q for Well1')
plt.title('Q for Well1')
plt.legend(loc='best', numpoints=1)
plt.show()

plt.plot( yit[1:iterations],Q[1][0:iterations-1], 'b.-',label='Q for Well2')
plt.title('Q for Well2')
plt.legend(loc='best', numpoints=1)
plt.show()

plt.plot( yit[1:iterations],Q[2][0:iterations-1], 'g.-',label='Q for Well3')
plt.title('Q for Well3')
plt.legend(loc='best', numpoints=1)
plt.show()
plt.plot( yit[1:iterations],Q[3][0:iterations-1], 'r.-',label='Q for Well4')
plt.title('Q for Well4')
plt.legend(loc='best', numpoints=1)
plt.show()
plt.plot( yit[1:iterations],Q[4][0:iterations-1], 'b.-',label='Q for Well5')
plt.title('Q for Well5')
plt.legend(loc='best', numpoints=1)
plt.show()

plt.plot( yit[1:iterations],Q[0][0:iterations-1], 'r-',label='Well1')
plt.plot(yit[1:iterations],Q[1][0:iterations-1], 'b--',label='Well2')
plt.plot(yit[1:iterations],Q[2][0:iterations-1], 'g-',label='Well3')
plt.plot(yit[1:iterations],Q[3][0:iterations-1], 'm--',label='Well4')
plt.plot(yit[1:iterations],Q[4][0:iterations-1], 'k-',label='Well5')
plt.title('Q for Wells')
plt.legend(loc='best', numpoints=1)
plt.show()

